from PyQt5.QtCore import Qt, QRect
from PyQt5.QtGui import QColor
from qgis.core import QgsWkbTypes, QgsFeatureRequest, QgsVectorLayer
from qgis.gui import QgsMapTool, QgsRubberBand
import sip

# Python adaptation of app/qgsmaptoolselectrectangle.cpp

def _set_rubber_band(canvas, select_rect, rubber_band):
    tr = canvas.getCoordinateTransform()
    ll = tr.toMapCoordinates(select_rect.left(), select_rect.bottom())
    lr = tr.toMapCoordinates(select_rect.right(), select_rect.bottom())
    ul = tr.toMapCoordinates(select_rect.left(), select_rect.top())
    ur = tr.toMapCoordinates(select_rect.right(), select_rect.top())
    if rubber_band:
        rubber_band.reset(QgsWkbTypes.PolygonGeometry)
        rubber_band.addPoint(ll, False)
        rubber_band.addPoint(lr, False)
        rubber_band.addPoint(ur, False)
        rubber_band.addPoint(ul, True)

def _expand_select_rectangle(select_rect, layer, point):
    box_size = 0
    if layer.geometryType() != QgsWkbTypes.PolygonGeometry:
        box_size = 5
    else:
        box_size = 1
    select_rect.setLeft(point.x() - box_size)
    select_rect.setRight(point.x() + box_size)
    select_rect.setTop(point.y() - box_size)
    select_rect.setBottom(point.y() + box_size)

def _get_matching_features(canvas, select_geometry, do_contains, single_select, layer):
    new_selected_features = []
    if select_geometry.type() != QgsWkbTypes.PolygonGeometry:
        return new_selected_features

    found_single_feature = False
    closest_id = 0
    closest_dist = None
    request = QgsFeatureRequest()
    request.setFilterRect(select_geometry.boundingBox())
    request.setFlags(QgsFeatureRequest.ExactIntersect)
    request.setSubsetOfAttributes([])
    for f in layer.getFeatures(request):
        g = f.geometry()
        if do_contains and not select_geometry.contains(g):
            continue
        elif not do_contains and not select_geometry.intersects(g):
            continue
        if single_select:
            found_single_feature = True
            distance = g.distance(select_geometry)
            if closest_dist is None or distance <= closest_dist:
                closest_dist = distance
                closest_id = f.id()
        else:
            new_selected_features.append(f.id())
    if single_select and found_single_feature:
        new_selected_features.append(closest_id)
    return new_selected_features

def _select_single_feature(canvas, select_geometry, e, layer):
    selected_features = _get_matching_features(canvas, select_geometry, False, True, layer)
    if selected_features == []:
        if not (e.modifiers() & Qt.ShiftModifier or e.modifiers() & Qt.ControlModifier):
            layer.removeSelection()
        return

    behavior = QgsVectorLayer.SetSelection

    if e.modifiers() & Qt.ShiftModifier or e.modifiers() & Qt.ControlModifier:
        select_id = selected_features[0]
        layer_selected_features = layer.selectedFeatureIds()
        if select_id in layer_selected_features:
            behavior = QgsVectorLayer.RemoveFromSelection
        else:
            behavior = QgsVectorLayer.AddToSelection

    layer.selectByIds(selected_features, behavior)

def _select_multiple_features(canvas, select_geometry, e, layer):
    behavior = QgsVectorLayer.SetSelection
    if e.modifiers() & Qt.ShiftModifier or e.modifiers() & Qt.ControlModifier:
        behavior = QgsVectorLayer.IntersectSelection
    elif e.modifiers() & Qt.ShiftModifier:
        behavior = QgsVectorLayer.AddToSelection
    elif e.modifiers() & Qt.ControlModifier:
        behavior = QgsVectorLayer.RemoveFromSelection

    selected_features = _get_matching_features(canvas, select_geometry, e.modifiers() & Qt.AltModifier, False, layer)
    layer.selectByIds(selected_features)

class MapToolSelectRectangle(QgsMapTool):
    def __init__(self, mapcanvas, layers):
        """
        :param QgsMapCanvas mapcanvas: a mapcanvas where to enable the maptool
        :param [QgsVectorLayer] layers: list of layers where to select on
        """
        QgsMapTool.__init__(self, mapcanvas)
        self._rubber_band = None
        self._fill_color = QColor(254, 178, 76, 63)
        self._stroke_color = QColor(254, 58, 29, 100)
        self._dragging = False
        self._layers = layers
        self._select_rect = QRect()

    def canvasPressEvent(self, e):
        self._select_rect.setRect(0, 0, 0, 0)
        if self._rubber_band is not None:
            sip.delete(self._rubber_band)
        self._rubber_band = QgsRubberBand(self.canvas(), QgsWkbTypes.PolygonGeometry)
        self._rubber_band.setFillColor(self._fill_color)
        self._rubber_band.setStrokeColor(self._stroke_color)

    def canvasMoveEvent(self, e):
        if e.buttons() != Qt.LeftButton:
            return

        if not self._dragging:
            self._dragging = True
            self._select_rect.setTopLeft(e.pos())
        self._select_rect.setBottomRight(e.pos())

        _set_rubber_band(self.canvas(), self._select_rect, self._rubber_band)

    def _select_on_layer(self, e, layer):
        select_rect = QRect(self._select_rect)
        if not self._dragging:
            _expand_select_rectangle(select_rect, layer, e.pos())
        else:
            if self._select_rect.width() == 1:
                self._select_rect.setLeft(select_rect.left()+1)
            if self._select_rect.height() == 1:
                self._select_rect.setBottom(select_rect.bottom()+1)
    
        if self._rubber_band:
            _set_rubber_band(self.canvas(), select_rect, self._rubber_band)

            select_geom = self._rubber_band.asGeometry()
            if not self._dragging:
                _select_single_feature(self.canvas(), select_geom, e, layer)
            else:
                _select_multiple_features(self.canvas(), select_geom, e, layer)

    def canvasReleaseEvent(self, e):
        for layer in self._layers:
            self._select_on_layer(e, layer)

        if self._rubber_band:
            sip.delete(self._rubber_band)
            self._rubber_band = None

        self._dragging = False



